#include <slb_common.h>


/* Defines */
#define SLB_DOWN	0
#define SLB_UP		1


/* Global variables */
static unsigned short	debug = 10;
static int		become_daemon = 1;
static int		nice_value = -19;

/* Internal variables */
struct Pool
{
	struct Pool		*prev;
	struct Pool		*next;

	unsigned int		old_weight;
	unsigned int		weight;
	unsigned long long	id;
	unsigned int		status;

	char			slave_ip[64];
	int			slave_port;

	char			service_ip[64];
	int			service_port;
	int			family;
	int			protocol;

	unsigned int		set_weight_counter;
	int			in_kernel;
	int			port_busy;

	char			*before, *after; /* used for [ and ] */
};

static struct Pool	*HEAD = NULL;
static struct Pool	*TAIL = NULL;
static FILE		*Logf = NULL;
static char		*log_file = "/var/log/slb/slbmd.log";
static int		port = 9100;
static short		ipv4 = 1, ipv6 = 1;


/* ip_vs stuff */

/*
 * Low level set of weight
 */
static void ip_vs_set_weight(struct Pool *p, unsigned int weight)
{
	char cmd[1024], *protocol;

	p->old_weight = p->weight;
	p->weight = weight;

	if (p->weight == p->old_weight)
		return;

	Log(4, "\tSet weight to %u on pool id %llu (%s/%d).\n",
		p->weight, p->id, p->slave_ip, p->slave_port);

	if (p->protocol == SOCK_DGRAM)
		protocol = "-u";
	else
		protocol = "-t";

	snprintf(cmd, sizeof(cmd), "/sbin/ipvsadm -e %s %s%s%s:%d"
		" -r %s%s%s:%d -m -w %u",
		protocol,
		p->before, p->service_ip, p->after,
		p->service_port,
		p->before, p->slave_ip, p->after,
		p->slave_port, weight);
	slb_run(cmd);
}

/*
 * Add a service
 */
static void ip_vs_add_service(char *ip, int port, const int family,
	const int protocol)
{
	char cmd[1024], cmd_protocol[16];
	char *before = "", *after = "";

	Log(1, "Add service family=%s/%s/%d/%s...\n",
		slb_get_family_type(family), ip, port,
		slb_get_protocol(protocol));

	if (protocol == SOCK_DGRAM)
		strcpy(cmd_protocol, "-u");
	else
		strcpy(cmd_protocol, "-t");

	if (family == AF_INET6) {
		before = "[";
		after = "]";
	}

	/* Remove first */
	snprintf(cmd, sizeof(cmd), "/sbin/ipvsadm -D %s %s%s%s:%d",
		cmd_protocol, before, ip, after, port);
	slb_run(cmd);

	/* Add */
	/* TODO: add -p? */
	snprintf(cmd, sizeof(cmd), "/sbin/ipvsadm -A %s %s%s%s:%d -s wlc",
		cmd_protocol, before, ip, after, port);
	slb_run(cmd);
}

/*
 * Add an ip to the pool
 */
static void ip_vs_add_slave(struct Pool *p)
{
	int ret;
	char cmd[1024], *protocol;

	if (!p) {
		Log(0, "%s: BUG: p is NULL!\n",
			__FUNCTION__);
		return;
	}

	p->set_weight_counter = 1;
	p->status = SLB_UP;
	p->port_busy = 1;

	if (p->in_kernel == 0) {
		if (p->protocol == SOCK_DGRAM)
			protocol = "-u";
		else
			protocol = "-t";
		snprintf(cmd, sizeof(cmd), "/sbin/ipvsadm -a %s %s%s%s:%d"
			" -r %s%s%s:%d -m -w 1",
			protocol,
			p->before, p->service_ip, p->after,
			p->service_port,
			p->before, p->slave_ip, p->after,
			p->slave_port);
		ret = slb_run(cmd);
		if (ret == 0)
			p->in_kernel = 1;
	}
}

/*
 * Remove an ip to the pool
 */
static void ip_vs_del_slave(struct Pool *p)
{
	if (!p) {
		Log(0, "%s: BUG: p is NULL!\n",
			__FUNCTION__);
		return;
	}

	p->status = SLB_DOWN;

	ip_vs_set_weight(p, 0);
}


/* Pool stuff */
static void Pool_dump(void)
{
	struct Pool *p;

	Log(6, "***** DUMP POOL: HEAD=%p TAIL=%p\n",
		HEAD, TAIL);
	p = HEAD;
	while (p) {
		Log(6, "\t%s service=%s/%s/%lu/%s slave=%s/%d id=%llu weight=%u"
			" in_kernel=%d port_busy=%d\n",
			p->status != SLB_UP ? "DOWN" : "UP",
			slb_get_family_type(p->family),
			p->service_ip, p->service_port,
			slb_get_protocol(p->protocol),
			p->slave_ip, p->slave_port, p->id, p->weight,
			p->in_kernel, p->port_busy);

		p = p->next;
	}

	Log(6, "***** DUMP POOL STOP\n");
}

static void Pool_add(const char *service_ip, const int service_port,
	const int family, const int protocol, const char *slave,
	const int slave_port, struct Conn *C)
{
	struct Pool *p;
	char *slave_ip;

	Log(1, "%s: Add slave %s/%d on Conn_id %llu to service"
		" family=%s/%s/%d/%s...\n",
		__FUNCTION__, slave, slave_port, Conn_getid(C),
		slb_get_family_type(family), service_ip, service_port,
		slb_get_protocol(protocol));

	p = malloc(sizeof(struct Pool));
	if (!p) {
		Log(0, "FATAL: Cannot alloc memory! Close conn!");
		Conn_close(C);
		return;
	}

	memset(p, 0, sizeof(struct Pool));

	p->prev = NULL;
	p->next = NULL;
	p->id = Conn_getid(C);
	p->set_weight_counter = 1;
	p->status = SLB_DOWN;
	slave_ip = Conn_addr_remote(C);
	snprintf(p->slave_ip, sizeof(p->slave_ip), "%s", slave_ip);
	p->slave_port = slave_port;
	snprintf(p->service_ip, sizeof(p->service_ip), "%s", service_ip);
	p->service_port = service_port;
	p->family = family;
	p->protocol = protocol;

	if (p->family == AF_INET6) {
		p->before = "[";
		p->after = "]";
	} else {
		p->before = "";
		p->after = "";
	}

	if (!TAIL) {
		HEAD = p;
		TAIL = p;
	} else {
		TAIL->next = p;
		p->prev = TAIL;
		TAIL = p;
	}
}

static struct Pool *Pool_get(const unsigned long long id)
{
	struct Pool *p, *ret = NULL;

	p = HEAD;
	while (p) {
		if (p->id == id) {
			ret = p;
			break;
		}

		p = p->next;
	}

	return ret;
}

static void s_trigger(struct Conn *C)
{
	struct Pool *p;
	char buf[1024];

	Log(4, "\tTrigger on id %llu...\n", Conn_getid(C));

	snprintf(buf, sizeof(buf), "LOAD\n");
	Conn_enqueue(C, buf, strlen(buf));

	p = Pool_get(Conn_getid(C));
	snprintf(buf, sizeof(buf),
		"PORT addr=%s protocol=%s port=%d\n",
		p->slave_ip, slb_get_protocol(p->protocol), p->slave_port);
	Conn_enqueue(C, buf, strlen(buf));

	/* don't trigger before we receive an response */
	Conn_set(C, CONN_PARA_TRIGGER, 0);
}

static void s_connected(struct Conn *C)
{
	struct Pool *p;

	Log(1, "%s: A connection was made to slave [%s/%d].\n",
		__FUNCTION__, Conn_addr_remote(C), Conn_port_remote(C));

	p = Pool_get(Conn_getid(C));
	ip_vs_add_slave(p);

	Conn_set(C, CONN_PARA_TRIGGER, 1);

	Pool_dump();
}

static void s_close(struct Conn *C)
{
	struct Pool *p;

	Log(1, "%s (A connection will be closed [%s] on slot %d)\n",
		__FUNCTION__, Conn_addr_remote(C), Conn_port_remote(C));

	p = Pool_get(Conn_getid(C));
	ip_vs_del_slave(p);

	Pool_dump();
}

static int s_data_cb(struct Conn *C, char *line)
{
	char *q;
	unsigned int line_size;
	char *dump;
	struct Pool *p;
	unsigned int load, weight, set_weight;
	unsigned long long cid;

	cid = Conn_getid(C);

	Log(3, "%s: cid=%llu...\n",
		__FUNCTION__, cid);

	p = Pool_get(cid);
	if (!p) {
		Log(0, "\tERROR: Invalid pool returned!\n");
		return -1;
	}

	line_size = strlen(line);
	if (line_size == 0)
		return 0;

	if (debug >= 8) {
		dump = Conn_dump(line, line_size);
		Log(0, "\tINFO: recv %u bytes: [%s]\n",
			line_size, dump);
		free(dump);
	}

	weight = 0;
	set_weight = 1;
	if (strncmp(line, "LOAD", 4) == 0) {
		Log(4, "\tLOAD answer received.\n");
		q = strstr(line, " load=");
		if (q) {
			load = strtol(q + strlen(" load="), NULL, 10);
			if (load >= 10000)
				weight = 1;
			else
				weight = 10000 - load;
		} else {
			Log(0, "\tWARN: Malformed LOAD line received [%s]!\n",
				line);
		}
	} else if (strncmp(line, "PORT", 4) == 0) {
		Log(4, "\tPORT answer received.\n");
		q = strstr(line, " busy=");
		if (q) {
			p->port_busy = strtoul(q + strlen(" busy="), NULL, 10);
			if (p->port_busy == 0) {
				Log(2, "\tWARN: slave do not listen! Set weight 0!\n");
				p->set_weight_counter = 1;
				weight = 0;
				/*TODO: run a command to send an alarm */
			} else {
				set_weight = 0;
			}
		} else {
			Log(0, "\tWARN: Malformed PORT line received [%s]!\n",
				line);
		}
	} else {
		Log(4, "\tOld style LOAD answer received. Please upgrade!\n");
		load = strtol(line, NULL, 0);
		if (load >= 10000)
			weight = 1;
		else
			weight = 10000 - load;
	}

	if (set_weight == 1) {
		p->set_weight_counter--;
		if (p->set_weight_counter == 0) {
			p->set_weight_counter = 5;

			ip_vs_set_weight(p, weight);
		}
	}

	Conn_set(C, CONN_PARA_TRIGGER, 1);

	return 0;
}

static void s_data(struct Conn *C)
{
	Conn_for_every_line(C, s_data_cb);
}

static void s_error(struct Conn *C)
{
	struct Pool *p;

	Log(5, "%s: cid=%llu [%s]\n",
		__FUNCTION__, Conn_getid(C),
		Conn_errno(C));

	p = Pool_get(Conn_getid(C));
	if (!p) {
		Log(0, "\tERROR: Invalid pool returned!\n");
		return;
	}

	ip_vs_set_weight(p, 0);
}

/*
 * Adds a service to the pool.
 * Returns 0 if ok, -1 on error
 * Format is:
 *	master_ip/master_port[/{tcp,udp}] slave1_ip/slave1_port ... slaveN_ip/slaveN_port
 */
static int add_line(char *s)
{
	char *p, *slaves, *slave, *slash, *master, *space, sprotocol[16], *next;
	int master_port, slave_port;
	struct Conn *C;
	int family, protocol;

	/* defaults */
	protocol = SOCK_STREAM;
	snprintf(sprotocol, sizeof(sprotocol), "tcp");

	/* find master_ip/master_port/protocol tokens */
	master = s;
	p = strchr(s, ' ');
	if (!p)
		return -1;
	*p = '\0';
	slaves = p + 1;

	/* slash will contain /80/tcp, for example */
	slash = strchr(master, '/');
	if (!slash) {
		/* Port not specified! */
		return -2;
	}
	*slash = '\0';
	/* next points to "80/tcp" */
	next = slash + 1;
	master_port = strtol(slash + 1, &next, 10);
	/* protocol can be tcp or udp for now */
	/* next is now at "/tcp" */
	if (*next != '\0') {
		next++; /* skip '/' */
		slash = strchr(next, '/');
		if (slash)
			*slash = '\0';
		snprintf(sprotocol, sizeof(sprotocol), "%s", next);
		if (strcmp(sprotocol, "udp") == 0)
			protocol = SOCK_DGRAM;
	}

	family = Conn_addr_family(master);
	if (family == -1)
		return -3;

	Log(1, "Found master %s/%d/%s/af=%s.\n",
		master, master_port, sprotocol, slb_get_family_type(family));

	ip_vs_add_service(master, master_port, family, protocol);

	/* now, the slaves */
	while (1) {
		Log(1, "Parsing slaves rest=[%s]...\n", slaves);
		if (slaves == NULL)
			break;

		slave = slaves;
		space = strchr(slaves, ' ');
		if (space) {
			*space = '\0';
			slaves = space + 1;
		} else {
			slaves = NULL;
		}

		/* Format: slave_addr/port */
		slash = strchr(slave, '/');
		if (!slash) {
			/* borrow port from master */
			slave_port = master_port;
		} else {
			/* Do not remove - finish slave part */
			*slash = '\0';
			slave_port = strtol(slash + 1, NULL, 10);
		}

		Log(1, "\tConnecting to slave %s/%s/%d/%s...\n",
			slb_get_family_type(family), slave, slave_port,
			slb_get_protocol(protocol));
		C = Conn_connect(family, protocol, slave, port);
		if (!C) {
			Log(1, "WARN: Cannot connect to %s/%d!"
				" I will try later.\n",
				slave, port);
		}
		Pool_add(master, master_port, family, protocol, slave,
			slave_port, C);

		Conn_set_cb(C, CONN_CB_TRIGGER, s_trigger);

		Conn_set(C, CONN_PARA_AUTO_RECONNECT, 1);
		Conn_set(C, CONN_PARA_RECONNECT_DELAY, 5);

		Conn_set(C, CONN_PARA_READ_TIMEOUT, 2000);
	}

	return 0;
}

int main(void)
{
	int ret;
	struct timeval start;
	FILE *f;
	char buf[4096], *r;

	Logf = fopen(log_file, "a+");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			log_file, strerror(errno));
		return 1;
	}
	setlinebuf(Logf);

	ret = Conn_init(0);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	Conn_debug(Logf, debug);

	gettimeofday(&start, NULL);
	Log(0, "Starting at %ld...\n", start.tv_sec);
	Log(0, "\tPort=%d\n", port);
	Log(0, "\tipv4=%s ipv6=%s\n", (ipv4 == 1) ? "on" : "off", (ipv6 == 1) ? "on" : "off");

	f = fopen("/etc/slb/slbmd.conf", "r");
	if (!f) {
		Log(0, "\tCannot open /etc/slb/slbmd.conf file (%s)!\n",
			strerror(errno));
		return 1;
	}

	while (fgets(buf, sizeof(buf) - 1, f)) {
		r = buf + strlen(buf) - 1;
		while ((r >= buf) && ((*r == ' ') || (*r == '\t') || (*r == '\n') || (*r == '\r'))) {
			*r = '\0';
			r--;
		}

		if ((strlen(buf) == 0) || (buf[0] == '#') || (buf[0] == ';'))
			continue;

		if (strncmp(buf, "debug", 5) == 0) {
			r = buf + 5;
			while ((*r == ' ') || (*r == '\t') || (*r == '='))
				r++;
			debug = strtoul(r, NULL, 10);
			Log(0, "\tSetting debug level to %u.\n", debug);
			Conn_debug(Logf, debug);
			continue;
		}

		if (strncmp(buf, "daemon", 6) == 0) {
			r = buf + 6;
			while ((*r == ' ') || (*r == '\t') || (*r == '='))
				r++;
			become_daemon = strtol(r, NULL, 10);
			Log(0, "\tSetting daemon to %u.\n", become_daemon);
			continue;
		}

		if (strncmp(buf, "nice", 4) == 0) {
			r = buf + 4;
			while ((*r == ' ') || (*r == '\t') || (*r == '='))
				r++;
			nice_value = strtol(r, NULL, 10);
			Log(0, "\tSetting nice to %u.\n", nice_value);
			continue;
		}

		ret = add_line(buf);
		if (ret != 0) {
			Log(0, "Error parsing [%s]. Ignore line [code=%d]!\n",
				buf, ret);
			continue;
		}
	}
	fclose(f);

	if (nice(nice_value) == -1)
		Log(0, "\tCannot set nice to %d. Continue...\n", nice_value);

	if (become_daemon == 1)
		daemon(0, 0);

	Pool_dump();

	Conn_connected_cb = s_connected;
	Conn_data_cb = s_data;
	Conn_close_cb = s_close;
	Conn_error_cb = s_error;

	while (1) {
		ret = Conn_poll(-1);
		if (ret == -1) {
			Log(0, "Error calling Conn_poll [%s]!\n", Conn_strerror());
			return 1;
		}
	}

	Log(0, "Finish!\n\n");

	return 0;
}
